
#define SPIN__SOURCE 1

lib_public {

    // spin.l                                    spin geometry in a window
    // by Mark Duchaineau (free but copyrighted, see LibGen/COPYING)
    //
    // 06-07-99: wrote
    // 10-15-99: made into a library module
    // 12-17-00: added _setmodelvmat()
    // 12-23-00: actually use name in spin_create()
    // 01-30-01: new 'a'/'z' key interpretation: acceleration
    // 01-31-01: moved 'w' screen dump keystroke to be common for all apps;
    //           added 'f'/'F' for frames-per-second meter
    // 03-28-01: eliminated possibility of drift in fps numerics
    // 04-07-01: made application-accessible variable spin_fps
    // 01-13-02: store final view matrix vm
    // 03-08-02: use of _perqz, added _setper()
    // 03-30-02: _setmsg
    // 05-08-02: _noclearswap,_clear,_swap
    // 05-17-02: put fps variable in spin record
    // 07-03-02: separate flags for NOCLEAR and NOSWAP in _noclearswap()
    // 07-11-02: simple menu display, default menu w/spin keys
    // 07-13-02: menu system hierarchies, submenu toggle
    // 07-14-02: completed initial menu system


    lib_include "Geowin/geowin.l"; // for win, view and 4x4 view-matrix stuff


    // hardware-specific defines
    lib_configexec "./configsub";
    lib_cleanexec "./cleansub";
    lib_include "spindefs.h";


    #define SPIN_FPSN 100
    #define SPIN_FPSNF 100.0


    // argument to _noclearswap() should OR together these bit flags
    #define SPIN_NOCLEAR 1
    #define SPIN_NOSWAP  2


    // size of keycode-to-menu hash table (should be power of two)
    #define SPIN_KEYHASH 256


    // NOTE: application callback gets events very similar to win events,
    // except some are caught by spin and not passed on and the first
    // callback arg is a spin object.  Take eventnum usage from win.l.
    // Spin windows are automatically double-buffered and
    // win_repeat/norepeated, so don't try to use those commands.  It is
    // safest to only read window state and perform drawing commands,
    // and avoiding making changes to the win directly (use or create
    // spin_...() wrappers in general for actions modifying win state).

    typedef void (*spinevent)(spin sp,int eventnum);


    lib_struct {
        spinevent event;             // application callback for events
        char *data;                  // application-supplied attachment
        win w;                       // owned by spin
        view v;                      // owned by spin
        int dragmode;                // drag state
        int dragmode0;               //   previous
        double v_x0,v_y0,v_x1,v_y1;  // more drag state
        double speed;                // eyepoint speed
        double px,py,pz;             // eyepoint
        double q,z0,z1,z2;           // perspective parameters
        double px0,py0,pz0;          // start-drag eyepoint
        int mx0,my0;                 // start-drag mouse pos
        int ax0,ay0;                 // start loc for menu drag
        vmat mmodel,mmov,mrot,mper;  // view matrix components
        vmat mmodmov,mmrot,mview,vm; // view matrix aggregates
        str msg;                     // application-supplied display message
        int noclearswap;             // 0=auto clear+swap (default), see defs
        int firstdraw;               // application-supplied display message
        double t0,t1,dt;             // frame timeofday, old and current
        int ti;                      // current index into circular dttab[]
        char ts[256];                // stats display string
        double dttab[SPIN_FPSN];     // circular frametime array
        double dtsum;                // running sum
        double fps;                  // average frames/sec for recent frames
        pack pk_menus;               // all menus
        pack pk_spm_default;         // predefined, access via _defaultmenu()
        pack keyhash[SPIN_KEYHASH];  // hash lookup for keycode menus
        int draw_default;            // toggle on/off default menu draw
        pack pk_spm_pick;            // selected menu
        pack pk_spm_pick0;           //   previous one
        int incode,incode0;          //   part of menu occupied now, previous
        spinmenu spm;                // spinmenu that generated event, if any
    } spin;


    // for use by applications
    spin spin_create(char *name,int x,int y,int xsiz,int ysiz,spinevent event);
    void spin_destroy(spin sp);
    void spin_setmodelvmat(spin sp,vmat mmodel);
    void spin_setper(spin sp,double q,double z0,double z1,double z2);
    void spin_setmsg(spin sp,char *msg);
    pack spin_defaultmenu(spin sp);
    void spin_setnoclearswap(spin sp,int noclearswap);
    void spin_clear(spin sp);
    void spin_swap(spin sp);


    lib_archive {
        "Spinmenu/spinmenu.l"
    };

};


// ...more: need to make geobench not tied to nvidia OpenGL extensions
//lib_application "Geobench/geobench.l";
lib_application "Spintest/spintest.l";


extern Display *win_display;


void spin_wevent(win w,int e);
void spin_vevent(view v);
void spin_do_draw(spin sp,int mx,int my);
void spin_loadper(spin sp);


spin spin_create(char *name,int x,int y,int xsiz,int ysiz,spinevent event)
{
    spin sp;

    sp=(spin)mem_get(sizeof(spin_struct),"spin_create");
    sp->firstdraw=1;
    sp->noclearswap=0;
    sp->event=event;
    sp->data=(char *)0;
    win_doublebuffer();
    sp->w=win_create(name,x,y,xsiz,ysiz,spin_wevent);
    sp->w->data=(char *)sp;
    win_repeat();
    sp->v=view_create((double)sp->w->xsize/2.0,(double)sp->w->ysize/2.0,
        (double)(sp->w->xsize+sp->w->ysize)/4.0,spin_vevent);
    sp->dragmode=sp->dragmode0=0;
    sp->v_x0=0.0; sp->v_y0=0.0; sp->v_x1=0.0; sp->v_y1=0.0;
    sp->speed=0.0;
    sp->px=0.0; sp->py=2.0; sp->pz=0.0;
    sp->q=2.0; sp->z0=0.0; sp->z1=0.001; sp->z2=10.0;
    sp->mmodmov=vmat_create(); sp->mmov=vmat_create(); sp->mrot=vmat_create();
    sp->mper=vmat_create(); sp->mmrot=vmat_create(); sp->mview=vmat_create();
    sp->mmodel=vmat_create(); vmat_ident(sp->mmodel);
    sp->vm=vmat_create();
    sp->msg=str_create("",-1);
    sp->fps=1.0;
    {
        int h;
        pack pk;

        sp->pk_spm_pick0=sp->pk_spm_pick=(pack)0;
        sp->incode=sp->incode0=SPINMENU_INCODE_NOTIN;
        for (h=0;h<SPIN_KEYHASH;h++) sp->keyhash[h]=pack_create("h");
        sp->pk_menus=pack_create("menus");
        sp->pk_spm_default=pk=spinmenu_createroot(sp,"spin",(spinmenuevent)0);
        spinmenu_createkey(pk,'a',"a     -- accelerate");
        spinmenu_createkey(pk,'z',"z     -- decelerate");
        spinmenu_createkey(pk,' ',"space -- stop");
        spinmenu_createkey(pk,WIN_KEY_TAB,"tab   -- toggle menu display");
        spinmenu_createkey(pk,'w',"w     -- write screenshot");
        spinmenu_create(pk->pkc1,"name spinout.tif...",(spinmenuevent)0);
        spinmenu_createkey(pk,WIN_KEY_ESCAPE,"esc   -- quit");
        sp->draw_default=1;
        sp->spm=(spinmenu)0;
    }
    //glClearColor(0.0,0.0,0.0,1.0);
    glClearColor(0.2,0.5,0.9,1.0); // sky color for roamtest
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    #ifdef GL_MULTISAMPLE_SGIS
        glDisable(GL_MULTISAMPLE_SGIS);
    #endif
    glDisable(GL_LIGHTING);
    glShadeModel(GL_FLAT);
    //glDisable(GL_TEXTURE_2D);
    glDisable(GL_CULL_FACE);
    glDrawBuffer(GL_BACK);
    glViewport(0,0,sp->w->xsize,sp->w->ysize);
    return sp;
}

void spin_destroy(spin sp)
{
    int h;

    for (h=0;h<SPIN_KEYHASH;h++) pack_destroy(sp->keyhash[h]);
    pack_destroy(sp->pk_menus);
    vmat_destroy(sp->mmodmov); vmat_destroy(sp->mmov); vmat_destroy(sp->mrot);
    vmat_destroy(sp->mper); vmat_destroy(sp->mmrot); vmat_destroy(sp->mview);
    vmat_destroy(sp->mmodel); vmat_destroy(sp->vm);
    view_destroy(sp->v); win_destroy(sp->w);
    mem_put((char *)sp,"spin_destroy");
}

void spin_setmodelvmat(spin sp,vmat mmodel)
{
    vmat_copy(mmodel,sp->mmodel);
}

void spin_setper(spin sp,double q,double z0,double z1,double z2)
{
    sp->q=q; sp->z0=z0; sp->z1=z1; sp->z2=z2;
    spin_loadper(sp);
}

void spin_setmsg(spin sp,char *msg)
{
    str slist[3];

    if (sp->msg) str_destroy(sp->msg);
    if (msg) {
        slist[0]="  "; slist[1]=msg; slist[2]=(str)0;
        sp->msg=str_catlist(slist);
    }else sp->msg=str_create("",-1);
}


pack spin_defaultmenu(spin sp)
{
    return sp->pk_spm_default;
}


void spin_setnoclearswap(spin sp,int noclearswap)
{
    sp->noclearswap=noclearswap;
}

void spin_clear(spin sp)
{
    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
}

void spin_swap(spin sp)
{
    glXSwapBuffers(win_display,sp->w->impl->xwin);
    glDrawBuffer(GL_BACK);
}


// ========== internal routines ==========

void spin_wevent(win w,int e)
{
    int x,y;
    spin sp;
    double d;

    sp=(spin)w->data;
    win_mouse(&x,&y);
    switch (e) {
    case WIN_DRAW:
        x=(w->xsize+w->ysize)/2;
        view_resize((double)w->xsize/2.0,(double)w->ysize/2.0,(double)x/2.0);
        break;
    case WIN_RIGHTDOWN: // toggle menu on/off and drag it around
        if (sp->dragmode==0) {
            spinmenu spm;

            sp->dragmode=1;
            sp->mx0=x; sp->my0=y;
            if (sp->draw_default==0) {
                sp->draw_default=1;
                spinmenu_setloc(sp->pk_spm_default,x,y-w->ysize,
                    SPINMENU_LOCATE_UL,SPINMENU_LOCATE_UL);
            }else{
                if (sp->pk_spm_pick==(pack)0) sp->draw_default=0;
            }
            spinmenu_displaykids(sp->pk_spm_default,sp->draw_default);
            spm=(spinmenu)sp->pk_spm_default->data;
            sp->ax0=spm->xa; sp->ay0=spm->ya;
        }
        break;
    case WIN_RIGHTUP:
        if (sp->dragmode==1) { sp->dragmode=0; }
        break;
    case WIN_LEFTDOWN: // rotate view direction for flight
        if (sp->dragmode==0) {
            if (sp->pk_spm_pick!=(pack)0) {
                sp->dragmode=20;
            }else{
                sp->dragmode=2;
                view_startdragrot((double)w->xsize-(double)x,
                    (double)w->ysize-(double)y);
            }
        }
        break;
    case WIN_LEFTUP:
        switch (sp->dragmode) {
        case 2:
            sp->dragmode=0;
            break;
        case 20:
            sp->dragmode=0;
            if (sp->pk_spm_pick!=(pack)0 && sp->incode==SPINMENU_INCODE_STR) {
                spinmenu spm;

                // got activation of menu item, dispatch it
                spm=(spinmenu)sp->pk_spm_pick->data;
                if (spm->event!=(spinmenuevent)0)
                    (*spm->event)(spm,WIN_LEFTUP);
            }
            break;
        default:
            break;
        }
        break;
    case WIN_MIDDLEDOWN: // translate in plane of view
        if (sp->dragmode==0) {
            sp->dragmode=3;
            sp->px0=sp->px; sp->py0=sp->py; sp->pz0=sp->pz;
            sp->mx0=x; sp->my0=y;
        }
        break;
    case WIN_MIDDLEUP:
        if (sp->dragmode==3) { sp->dragmode=0; }
        break;
    case WIN_KEYDOWN:
// ...more: need to grab keyboard in win_ogl.l from keydown until keyup
        {
            int h;
            pack pk;
            spinkeycode spk;

            h=w->keycode&(SPIN_KEYHASH-1);
            for (pk=sp->keyhash[h]->pkc0;pk;pk=pk->pkl1) {
                spk=(spinkeycode)pk->data;
                if (spk->keycode==w->keycode) {
                    if (spk->spm->event!=(spinmenuevent)0)
                        (*spk->spm->event)(spk->spm,WIN_KEYDOWN);
                }
            }
        }
        break;
    case WIN_REPEAT:
        d=sp->speed*sp->dt*100.0;
        sp->px-=d*sp->v->vx; sp->py-=d*sp->v->vy; sp->pz-=d*sp->v->vz;
        switch (sp->dragmode) {
        case 1:
            if (sp->draw_default!=0) {
                int ax,ay;
                spinmenu spm;

                ax=sp->ax0+x-sp->mx0; ay=sp->ay0+y-sp->my0;
                spinmenu_setloc(sp->pk_spm_default,ax,ay,
                    SPINMENU_LOCATE_UL,SPINMENU_LOCATE_UL);
                spinmenu_drawprep(sp->pk_spm_default,x,y);
                spinmenu_fixloc(sp->pk_spm_default);
            }
            break;
        case 2:
            view_dragrot((double)w->xsize-(double)x,
                (double)w->ysize-(double)y);
            break;
        case 3:
            {
                double a,b;

                a=(double)(x-sp->mx0)/sp->v->scale;
                b=(double)(y-sp->my0)/sp->v->scale;
                sp->px=sp->px0+sp->v->ux*a+sp->v->wx*b;
                sp->py=sp->py0+sp->v->uy*a+sp->v->wy*b;
                sp->pz=sp->pz0+sp->v->uz*a+sp->v->wz*b;
            }
            break;
        // NOTE: case 20 handled in REPEAT draw handler
        default:
            break;
        }
/********************************************
 **  stuff for spin animation after leftup **
 ********************************************
        else{
            if (sp->leftup==1) {
                sp->v_x0=v->dragx; sp->v_y0=v->dragy;
                sp->v_x1=(double)x; sp->v_y1=(double)y;
                {
                    double dx,dy,r;

                    dx=sp->v_x1-sp->v_x0; dy=sp->v_y1-sp->v_y0;
                    r=sqrt(dx*dx+dy*dy);
                    sp->leftup=(r>1.9?2:0);
                }
            }
            if (sp->leftup==2) {
                v->dragx=sp->v_x0; v->dragy=sp->v_y0;
                view_dragrot(sp->v_x1,sp->v_y1);
            }
        }
********************************************/
        spin_do_draw(sp,x,y);
        break;
    default:
        sp->spm=(spinmenu)0;
        (*sp->event)(sp,e);
        break;
    }
}

void spin_vevent(view v)
{
    // do nothing (view changes are asyncronous to wevent activity)
}

void spin_do_draw(spin sp,int mx,int my)
{
    sp->firstdraw=0;

    spin_loadper(sp);

    if (!(sp->noclearswap&SPIN_NOCLEAR)) spin_clear(sp);

    sp->spm=(spinmenu)0;
    (*sp->event)(sp,WIN_DRAW);

    {
        int ti;
        double dt,dtsum;
        extern int win_font_tex_id[2];
        int x,y;

        sp->t1=timer_systemdsec();
        if (sp->t0!=0.0) {
            sp->dt=dt=sp->t1-sp->t0;
            if (sp->ti<0) {
                for (ti=0;ti<SPIN_FPSN;ti++) sp->dttab[ti]=dt;
                sp->ti=0; sp->dtsum=SPIN_FPSNF*dt;
            }
            sp->dtsum+=dt-sp->dttab[sp->ti]; sp->dttab[sp->ti]=dt; sp->ti++;
            if (sp->ti>=SPIN_FPSN) {
                for (ti=0,dtsum=0.0;ti<SPIN_FPSN;ti++) dtsum+=sp->dttab[ti];
                sp->dtsum=dtsum; sp->ti=0;
            }
            glLoadMatrixd(sp->w->impl->vm);
            glDisable(GL_DEPTH_TEST);
            sp->fps=1.0/(sp->dtsum/SPIN_FPSNF);
            sprintf(sp->ts,"fps: %7.3f%s",sp->fps,sp->msg);
            win_color(20,20,95);
            win_strsize(sp->ts,&x,&y);
            win_rectf(10,10,x,y);
            win_color(255,255,0);
            win_str(10,10,sp->ts);

            // draw all menus
            {
                pack pk;
                spinmenu spm;

                sp->pk_spm_pick0=sp->pk_spm_pick; sp->pk_spm_pick=(pack)0;
                sp->incode0=sp->incode; sp->incode=SPINMENU_INCODE_NOTIN;
                for (pk=sp->pk_menus->pkc0;pk;pk=pk->pkl1) {
                    spm=(spinmenu)pk->data;
                    if (spm->draw_kids!=0) {
                        spinmenu_drawprep(pk,mx,my);
                        if (spinmenu_fixloc(pk)!=0)
                            spinmenu_drawprep(pk,mx,my);
                    }
                }
                for (pk=sp->pk_menus->pkc0;pk;pk=pk->pkl1) {
                    spm=(spinmenu)pk->data;
                    if (spm->draw_kids!=0) spinmenu_draw(pk);
                }
                if (sp->dragmode==20 && sp->pk_spm_pick!=(pack)0 &&
                    (sp->pk_spm_pick!=sp->pk_spm_pick0 ||
                     sp->incode0!=sp->incode ||
                     sp->dragmode0!=sp->dragmode) &&
                    sp->incode==SPINMENU_INCODE_ROOT) {
                    spm=(spinmenu)sp->pk_spm_pick->data;
                    spm->draw_kids=(spm->draw_kids?0:1);
                }
                sp->dragmode0=sp->dragmode;
            }

            glEnable(GL_DEPTH_TEST);
        }
        sp->t0=sp->t1;
    }
    if (!(sp->noclearswap&SPIN_NOSWAP)) spin_swap(sp);
}


void spin_loadper(spin sp)
{
    int i,j,k;
    win w;
    double vm[16],qx,qy,q;
    view v;

    w=sp->w; v=sp->v;

    q=(double)(w->xsize>=w->ysize?w->xsize:w->ysize);
    qx=q/(double)w->xsize;
    qy=q/(double)w->ysize;

    // fly-around mode
    vmat_move(sp->px,sp->py,sp->pz,sp->mmov);
    vmat_view(v,sp->mrot);
    //vmat_per(4.0,sp->mper);
    vmat_perqz(sp->q,sp->z0,sp->z1,sp->z2,sp->mper);
    vmat_mul(sp->mmov,sp->mmodel,sp->mmodmov);
    vmat_mul(sp->mrot,sp->mmodmov,sp->mmrot);
    vmat_mul(sp->mper,sp->mmrot,sp->mview);

    // spin an object at the origin from (0,0,4) looking towards -z
    //vmat_view(v,sp->mrot);
    //vmat_move(0.0,0.0,4.0,sp->mmov);
    //vmat_per(4.0,sp->mper);
    //vmat_mul(sp->mmov,sp->mrot,sp->mmrot);
    //vmat_mul(sp->mper,sp->mmrot,sp->mview);

    if (!sp->firstdraw) {
        glMatrixMode(GL_PROJECTION);
        for (k=0;k<16;k++) {
            i=k%4; j=k/4;
            if (i==0) vm[k]=qx*sp->mview->m[i][j];
            else{
                if (i==1) vm[k]=qy*sp->mview->m[i][j];
                else vm[k]=sp->mview->m[i][j];
            }
            sp->vm->m[i][j]=vm[k];
        }
        glLoadMatrixd(vm);
    }
}


