
lib_public {

    // qpool.l        streamlined mempools (faster but fewer capabilities)
    // by Mark Duchaineau (free but copyrighted, see LibGen/COPYING)
    //
    // NOTE: these have a subset of the capabilities of mempools,
    // but are faster
    //
    // 08-30-96: outlined
    // 09-03-96: wrote
    // 04-02-97: moved into Sys, added _numactive()

    lib_include "../sys.l";

    #define QPOOL_N 256                   /* number of records per block */

    lib_struct {                          // global record for pool
        int size;                         // bytes per record
        qpoolref free,active;             // lists of unused and in-use records
        qpoolblock blocks;                // list of allocation blocks
    } qpool;

    lib_struct {                          // prepended to each pool record
        qpoolref qpr0,qpr1;               // links in ->free/->active list
        qpool qp;                         // pool that this is from
        int dummy;                        // added to keep 8-byte alignment
    } qpoolref;
    #define QPOOLREF_SIZE 16              /* == sizeof(qpoolref_struct) */

    lib_struct {                          // prepended to block of pool records
        qpoolblock qpb1;                  // link to next block
        int dummy;                        // added to keep 8-byte alignment
    } qpoolblock;
    #define QPOOLBLOCK_SIZE 8             /* == sizeof(qpoolblock_struct) */

    qpool qpool_create(int size);         // make new pool of size-byte records
    void qpool_destroy(qpool qp);         // get rid of pool
    char *qpool_get(qpool qp);            // grab another record to use
    void qpool_put(char *data);           // return record for later use
    int qpool_numactive(qpool qp);        // return count of active records

    // macros to loop through all active records
    // -----------------------------------------
    #define BGNQPOOL(QP) {                                                    \
        qpoolref qpr,qpr1;                                                    \
        char *qpdata;                                                         \
                                                                              \
        for (qpr=(QP)->active;qpr;qpr=qpr1) {                                 \
            qpr1=qpr->qpr1; qpdata=((char *)qpr)+QPOOLREF_SIZE;

    #define ENDQPOOL(QP)                                                      \
        }                                                                     \
    }

};


qpool qpool_create(int size)
{
    qpool qp;

    qp=(qpool)mem_get(sizeof(qpool_struct),"qpool_create");
    if (size&7) size=(size&~7)+8;
    qp->size=size;
    qp->free=(qpoolref)0; qp->active=(qpoolref)0;
    qp->blocks=(qpoolblock)0;
    return qp;
}

void qpool_destroy(qpool qp)
{
    qpoolblock qpb,qpb1;

    for (qpb=qp->blocks;qpb;qpb=qpb1) {
        qpb1=qpb->qpb1;
        mem_put((char *)qpb,"qpool_destroy");
    }
    mem_put((char *)qp,"qpool_destroy");
}

char *qpool_get(qpool qp)
{
    int i,dd;
    char *data;
    qpoolref qpr,qpr1;
    qpoolblock qpb;

    // grab first free pool record, adding another block if needed
    if (!(qpr=qp->free)) {
        dd=QPOOLREF_SIZE+qp->size;
        data=mem_get(QPOOLBLOCK_SIZE+QPOOL_N*dd,"qpool_get");
        qpb=(qpoolblock)data; qpb->qpb1=qp->blocks; qp->blocks=qpb;
        data+=QPOOLBLOCK_SIZE; qpr1=(qpoolref)0;
        for (i=0;i<QPOOL_N;i++,data+=dd)
            { qpr=(qpoolref)data; qpr->qp=qp; qpr->qpr1=qpr1; qpr1=qpr; }
        qp->free=qpr;
    }

    // move pool record from free list to active list
    qp->free=qpr->qpr1; qpr1=qp->active; qp->active=qpr;
    qpr->qpr0=(qpoolref)0; qpr->qpr1=qpr1; if (qpr1) qpr1->qpr0=qpr;

    // return application portion of pool record
    return ((char *)qpr)+QPOOLREF_SIZE;
}

void qpool_put(char *data)
{
    qpoolref qpr,qpr0,qpr1;
    qpool qp;

    // hmmm, this is dangerous code!
    // but speed is of the essence...

    // get qpoolref that is prepended to pool record,
    // and the global pool record from that
    qpr=(qpoolref)(data-QPOOLREF_SIZE); qp=qpr->qp;

    // remove record from active list
    qpr0=qpr->qpr0; qpr1=qpr->qpr1;
    if (qpr0) qpr0->qpr1=qpr1; else qp->active=qpr1;
    if (qpr1) qpr1->qpr0=qpr0;

    // insert record into free list
    qpr->qpr1=qp->free; qp->free=qpr;
}

int qpool_numactive(qpool qp)
{
    int n;
    qpoolref qpr;
    qpoolblock qpb;

    // the speed idea here: since the active records usually outnumber
    // the free ones, get a quick total and subtract the free ones
    for (n=0,qpb=qp->blocks;qpb;qpb=qpb->qpb1) n+=QPOOL_N;
    for (qpr=qp->free;qpr;qpr=qpr->qpr1) n--;
    return n;
}


